package xkpasswd;

use strict;

#################################################################
# Utility for generating XKCD/Password Haystacks-like passwords #
#################################################################

# Copyright 2011 Bart Busschots T/A Bartificer Web Solutions. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are
# permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this list of
#      conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this list
#      of conditions and the following disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY BART BUSSCHOTS T/A BARTIFICER WEB SOLUTIONS ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# BART BUSSCHOTS T/A BARTIFICER WEB SOLUTIONS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#
# config defaults
#
$xkpasswd::DICTIONARY_FILE = '/Users/bart/Documents/temp/fromSVN/xkpasswd/trunk/dict.txt'; # the location on the file
#                                                       system of the dictionary
#                                                       to use
$xkpasswd::MIN_WORD_LENGTH = 4; # the min length of words to be used
$xkpasswd::MAX_WORD_LENGTH = 8; # the max length of words to be used
$xkpasswd::MIN_WORDS = 100; # the minimum number of words in the dict to proceed
@xkpasswd::SEPARATORS = ('.', '*', '-', '_', ',', '#', '|', '%', '^', '&', '$', '=', '+');

#
# global variables
#
$xkpasswd::DEBUG = 0; # whether or not to debug

#
# constructur
#
# Arguments:
# 1. OPTIONAL - config file specifing the initialisation of the dictionary file.
#    That argument should be a hashref indexed by one or more of the keys below.
#    Any keys not specified will be interpreted as the defaults  defined above.
#    dictionary_file -> the file from which the dictionary words should be loaded.
#                       one word per line, lines starting with # are ignored.
#    min_word_length -> the minimum length of word that should be loaded into the
#                       dictionary from the file.
#    max_word_length -> the maximum length of word that should be loaded into the
#                       dictionary from the file.
# Returns: an object of type xkpasswd
sub new{
    my $class = shift;
    my $config = shift;
    my $self = {
                dictionary => []
                };
    bless $self, $class;
    $self->setDictionary($config->{'dictionary_file'}, $config->{'min_word_length'}, $config->{'max_word_length'});
    return $self;
}

#
# functions
#

# An instance method to update the dictionary settings of an xkpasswd instance.
# Any arguments which don't validate will be replaced with their defauly values.
# Arguments:
# 1. a dictionary file
# 2. the minimum length of the words to be accepted into the dictionary
# 3. the minimum length of the words to be accepted into the dictionary
# Returns: nothing
sub setDictionary{
    my $this = shift;
    my $dictionaryFile = shift;
    my $minWordLength = shift;
    my $maxWordLength = shift;
    
    # set the values (or defaults in case of error)
    if($dictionaryFile){
        $this->{'dictionary_file'} = $dictionaryFile;
    }else{
        $this->{'dictionary_file'} = $xkpasswd::DICTIONARY_FILE;
    }
    if($minWordLength && $minWordLength =~ m/^\d+$/ && $minWordLength > 0){
        $this->{'min_word_length'} = $minWordLength;
    }else{
        $this->{'min_word_length'} = $xkpasswd::MIN_WORD_LENGTH;
    }
    if($maxWordLength && $maxWordLength =~ m/^\d+$/ && $maxWordLength > 0){
        $this->{'max_word_length'} = $maxWordLength;
    }else{
        $this->{'max_word_length'} = $xkpasswd::MAX_WORD_LENGTH;
    }
    if($this->{'max_word_length'} < $this->{'min_word_length'}){
         $this->{'max_word_length'} =  $this->{'min_word_length'};
    }
    
    # blank the dictonary array ref
    $this->{'dictionary'} = [];
}

# An instance method to load the words of appropriate length from the dictionary file
# into the dictionary array ref in the object.
# Args: none
# Returns: 0 on error (did not load enough words), 1 on success
sub loadDict{
    my $this = shift;
    print "DEBUG - using dictionary file: ".$this->{'dictionary_file'}."\n" if $xkpasswd::DEBUG;
    
    open(DICT, $this->{'dictionary_file'}) || return 0;
    $this->{'dictionary'} = [];
    while(<DICT>){
        chomp;
        
        # skip 'comment' lines in the dictionary file (lines starting with a #)
        next if m/^#/;
        
        print "DEBUG - processing word $_\n" if $xkpasswd::DEBUG;
        
        if(length($_) >= $this->{'min_word_length'} && length($_) <= $this->{'max_word_length'}){
            push(@{$this->{'dictionary'}}, $_);
        }
    }
    close(DICT);
    if(scalar(@{$this->{'dictionary'}}) < $xkpasswd::MIN_WORDS){
        $this->{'dictionary'} = [];
        return 0;
    }
    return 1;
}

# A Class function to verify a config, and fill in any missing values with defaults
# Args: a hashref containing the config, config keys are:
#       num_words -> the number of words to use
#       l33t_substitutions -> a hashref of hashrefs indexed by the characters to be replaced, and
#                             containing the replacement strings. The substitutions are applied
#                             only to the words in the password, not the separators, padding, or
#                             numbers. If not set, no substitutions will be applied.
#       case_transform -> a value of NONE (leave the case as it came from the dictionary),
#                         UPPER (convert all words to all upper case), LOWER (convert all words
#                         to all lower case), CAPITAL (capitalise the first letter of the work
#                         convert all others to lower case), or RANDOM (have each words randomly
#                         all upper or lower case). Defaults to NONE.
#       custom_separator -> custom separator to override default list
#       preppend_numbers -> the number of digits to pre-pend the list of words
#       append_numbers -> the number of digits to append to the end of the list of words
#       pad_char -> the character to use for padding front and back (SEPARATOR to use the separator)
#       pre_pad -> the ammount of chars to pad at the start
#       post_pad -> the number of chars to pad at the end
#       adaptive_padding -> the total number of characters to padd the string out to
#                           if set, overrides pre and post padding numbers, and truncates
#                           the password to this length if the generated password is longer
#                           than this number.
#                           All adaptive padding is added to the end of the password.
#                           Dissable by setting to 0.

# Returns: a fully populated version of the config
sub populateConfig($){
    my $config = shift();
    
    # check the number of words
    unless($config->{'num_words'} =~ m/^\d+$/ && $config->{'num_words'} >= 1){
        $config->{'num_words'} = 4;
    }
    
    # check the case tranform
    unless($config->{'case_transform'} eq 'NONE' || $config->{'case_transform'} eq 'UPPER' || $config->{'case_transform'} eq 'LOWER' || $config->{'case_transform'} eq 'CAPITAL' || $config->{'case_transform'} eq 'RANDOM'){
        $config->{'case_transform'} = 'NONE';
    }
    
    # check the number of digits to prepend and append
    unless($config->{'preppend_numbers'} =~ m/^\d+$/ && $config->{'preppend_numbers'} >= 0){
        $config->{'preppend_numbers'} = 3;
    }
    unless($config->{'append_numbers'} =~ m/^\d+$/ && $config->{'append_numbers'} >= 0){
        $config->{'append_numbers'} = 3;
    }
    
    # check the padding parameters
    unless(length($config->{'pad_char'}) == 1 || $config->{'pad_char'} eq 'SEPARATOR'){
        $config->{'pad_char'} = 'SEPARATOR';
    }
    unless($config->{'pre_pad'} =~ m/^\d+$/ && $config->{'pre_pad'} >= 0){
        $config->{'pre_pad'} = 5;
    }
    unless($config->{'post_pad'} =~ m/^\d+$/ && $config->{'post_pad'} >= 0){
        $config->{'post_pad'} = 5;
    }
    unless($config->{'adaptive_padding'} =~ m/^\d+$/ && $config->{'adaptive_padding'} > 0){
        $config->{'adaptive_padding'} = 0;
    }
    
    return $config;
}

# An instance function to generate a password
# Args: optionally a hashref containing configuration keys
# Returns: 0 on error, a random password on success
sub generatePassword{
    my $this = shift;
    # get the passed config, if any
    my $config = shift;
    
    # populate defaults to fill any gaps in the passed config
    $config = populateConfig($config);
    
    #  load the dictionary if needed
    unless(scalar(@{$this->{'dictionary'}})){
        return 0 unless $this->loadDict();
    }
    
    # get the size of the dictionary
    my $dictLength = scalar(@{$this->{'dictionary'}});
    
    # gather the random words
    my @randomWords;
    while(scalar(@randomWords) < $config->{'num_words'}){
        my $randWord = $this->{'dictionary'}->[int(rand($dictLength))];
        
        # apply case transforms if needed
        if($config->{'case_transform'} eq 'UPPER'){
            $randWord = uc($randWord);
        }elsif($config->{'case_transform'} eq 'LOWER'){
            $randWord = lc($randWord);
        }elsif($config->{'case_transform'} eq 'CAPITAL'){
            $randWord = lc($randWord);
            $randWord = ucfirst($randWord);
        }elsif($config->{'case_transform'} eq 'RANDOM'){
            if(int(rand(1000)) % 2 == 0){
                $randWord = lc($randWord);
            }else{
                $randWord = uc($randWord);
            }
        }
        
        
        # apply 133t substitutions if needed
        if($config->{'l33t_substitutions'}){
            foreach my $key (keys(%{$config->{'l33t_substitutions'}})){
                my $sub = $config->{'l33t_substitutions'}->{$key};
                if($sub){
                    $randWord =~ s/$key/$sub/gi;
                }
            }
        }
        
        push(@randomWords, $randWord);
    }
    
    # get the separator
    my $separator = $config->{'custom_separator'};
    $separator = $xkpasswd::SEPARATORS[rand(scalar(@xkpasswd::SEPARATORS))] unless $separator;
    
    my $password = '';
    
    # prepend a number if required
    for(my $n=0; $n < $config->{'preppend_numbers'}; $n++){
        if($n == 0){
            $password .= int(rand(9))+1;
        }else{
            $password .= int(rand(10));
        }
    }
    $password .= $separator if $config->{'preppend_numbers'};
    
    # add the words
    $password .= join($separator, @randomWords);
    
    # append a number if required
    $password .= $separator if $config->{'append_numbers'};
    for(my $n=0; $n < $config->{'append_numbers'}; $n++){
        if($n == 0){
            $password .= int(rand(9))+1;
        }else{
            $password .= int(rand(10));
        }
    }
    
    # deal with the padding
    my $padChar = $separator;
    $padChar = $config->{'pad_char'} unless $config->{'pad_char'} eq 'SEPARATOR';
    if($config->{'adaptive_padding'}){
        # adaptive padding
        
        if(length($password) > $config->{'adaptive_padding'}){
            # trim back to adaptive padding length
            $password = substr($password, 0, $config->{'adaptive_padding'});
        }elsif(length($password) < $config->{'adaptive_padding'}){
            # pad the back of the password
            while(length($password) < $config->{'adaptive_padding'}){
                $password .= $padChar;
            }
        }
    }else{
        # regular padding
        for(my $n=0; $n < $config->{'pre_pad'}; $n++){
            $password = $padChar.$password;
        }
        for(my $n=0; $n < $config->{'post_pad'}; $n++){
            $password .= $padChar;
        }
    }
    
    return $password;
}

my $perl_is_dumb = 1;